//Microchip Solutions v2011-12-05
/******************************************************************************
            USB Custom Demo, Host

This file provides the main entry point to the Microchip USB Custom
Host demo.  This demo shows how a PIC24F system could be used to
act as the host, controlling a USB device running the Microchip Custom
Device demo.

 ******************************************************************************/

/******************************************************************************
 * Filename:        main.c
 * Dependancies:    USB Host Driver with Generic Client Driver
 * Processor:       PIC24F256GB1xx
 * Hardware:        Explorer 16 with USB PICtail Plus
 * Compiler:        C30 v2.01/C32 v0.00.18
 * Company:         Microchip Technology, Inc.

Software License Agreement

The software supplied herewith by Microchip Technology Incorporated
(the ?ompany? for its PICmicro? Microcontroller is intended and
supplied to you, the Company? customer, for use solely and
exclusively on Microchip PICmicro Microcontroller products. The
software is owned by the Company and/or its supplier, and is
protected under applicable copyright laws. All rights are reserved.
Any use in violation of the foregoing restrictions may subject the
user to criminal sanctions under applicable laws, as well as to
civil liability for the breach of the terms and conditions of this
license.

THIS SOFTWARE IS PROVIDED IN AN ?S IS?CONDITION. NO WARRANTIES,
WHETHER EXPRESS, IMPLIED OR STATUTORY, INCLUDING, BUT NOT LIMITED
TO, IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE APPLY TO THIS SOFTWARE. THE COMPANY SHALL NOT,
IN ANY CIRCUMSTANCES, BE LIABLE FOR SPECIAL, INCIDENTAL OR
CONSEQUENTIAL DAMAGES, FOR ANY REASON WHATSOEVER.


 *******************************************************************************/
//0x0004 = Attribute Protocol
//0x0005 = LE L2CAP Signaling channel
//0x0006 = Security Manager Protocol

#include "mcc_generated_files/system.h"
#include "mcc_generated_files/usb/usb.h"
#include "mcc_generated_files/uart2.h"
#include "mcc_generated_files/led.h"

// *****************************************************************************
// *****************************************************************************
// Internal Function Prototypes
// *****************************************************************************
// *****************************************************************************

void DelayMs(uint16_t);

// *****************************************************************************
// *****************************************************************************
// Macros
// *****************************************************************************
// *****************************************************************************

//#define DEBUG_MODE

#define MAX_NO_OF_BYTES  64

#define DISCONNECT 0x05
#define ENCRYPT_EVENT 0x08
//#define NUM_OF_COMP_PACKES_EVENT 0x13
#define BL_EVENT 0x0e
#define BLE_EVENT 0x3e

// *****************************************************************************
// *****************************************************************************
// Type definitions
// *****************************************************************************
// *****************************************************************************


// Application States

typedef enum {
    BT_INITIALIZE = 0, // Initialize the app when a device is attached
    BT_STATE_IDLE, // Inactive State
    BT_STATE_PROCESS,
    BT_STATE_WRITE_CLASS,
    BT_STATE_READ_EP1,
    BT_STATE_READ_CLASS_WAITING,
    BT_STATE_WRITE_ACL,
    BT_STATE_READ_ACL_HCI,
    BT_STATE_READ_ACL_WAITING,
    BT_STATE_READ_HCI_WAITING,
    BT_STATE_READ_HCI,

    BT_STATE_ERROR // An error has occured
} BT_STATE;

// Hci States

typedef enum {
    HCI_CMD_RESET = 0, // Initialize the hci when a device is attached
    HCI_CMD_RESET_END,
    HCI_CMD_READ_BD_ADDR,
    HCI_CMD_READ_BD_ADDR_END,
    HCI_CMD_MASK,
    HCI_CMD_MASK_END,
    HCI_CMD_SCAN_PARAM,
    HCI_CMD_SCAN_PARAM_END,
    HCI_CMD_SCAN_ENABLE,
    HCI_CMD_SCAN_ENABLE_END,
    HCI_CMD_SCAN_ENABLE_END1,
    HCI_CMD_CREATE_CONNECT,
    HCI_CMD_CLEAR_LIST,
    HCI_CMD_CLEAR_LIST_END,
    HCI_CMD_WRITE_LIST,
    HCI_CMD_WRITE_LIST_END,
    HCI_CMD_SCAN_DISABLE,
    HCI_CMD_SCAN_DISABLE_END,
    HCI_CMD_CREATE_CONNECT_END,

    L2CAP_DISCONNECT_REQ,
    L2CAP_DISCONNECT_REQ_END,

    SMP_SECURITY_REQ,
    SMP_PAIRING_REQ,
    SMP_PAIR_RES,
    SMP_FUNC_E,
    SMP_FUNC_E_END,
    SMP_FUNC_C1,
    SMP_FUNC_C1_END,
    SMP_FUNC_S1,
    SMP_FUNC_S1_END,
    SMP_PAIR_CONF_IN,
    SMP_PAIR_CONF_OUT,
    SMP_RANDOM_OUT,
    SMP_RANDOM_IN,
    SMP_LTK_REQ,
    SMP_LTK_RES,
    SMP_ENCRYP_INFO_IN,
    SMP_RANDEDIV_IN,
    SMP_ENCRYP_INFO_OUT,
    SMP_RANDEDIV_OUT,

    GATT_MTU_REQ,
    GATT_MTU_RES,
    GATT_GEN_ATT_CCC,
    GATT_GEN_ATT_CCC_RES,
    GATT_GEN_HID1_CCC,
    GATT_GEN_ATT_HID1_RES,
    GATT_GEN_HID2_CCC,
    GATT_GEN_ATT_HID2_RES,

    READ_HID_DATA,
    READ_HID_DATA1,
    READ_HID_DATA2,

    PROG_END

} HCI_STATE;


// *****************************************************************************
// *****************************************************************************
// Local Variables
// *****************************************************************************
// *****************************************************************************

uint8_t deviceAddress; // Address of the device on the USB
BT_STATE DemoState; // Current state of the demo application
HCI_STATE HciState; // Current state of the demo application
uint16_t data_size;
uint16_t data_num;

uint8_t buf[MAX_NO_OF_BYTES];
uint8_t buf1[MAX_NO_OF_BYTES];
uint8_t handle[2]; //a handle for ACL 
uint8_t hid_flag = 0;
uint8_t EP2busy_flag = 0;

uint8_t HID17_flag = 0;

uint16_t step;
char message[30];
uint16_t end_num;


uint8_t PAIR_CONF[16];
uint8_t PAIR_RANDOM[16];
uint8_t INIT_RANDOM[16] = {0x94, 0x58, 0x7f, 0x05, 0xdc, 0x5d, 0x8f, 0xe6, 0xdd, 0x26, 0x3e, 0x6d, 0x2b, 0x16, 0xb4, 0x7a};
uint8_t STK[16];

uint8_t INI_ADDRESS_TYPE = 0, RES_ADDRESS_TYPE = 0;
uint8_t INI_BT_ADDRESS[6], RES_BT_ADDRESS[6];
uint8_t PAIR_RES[7], PAIR_REQ[7] = {0x01, 0x04, 0x00, 0x2d, 0x10, 0x0e, 0x0f};

uint8_t P1[16], P2[16];

uint8_t IRK[16] = {0xaa, 0xaa, 0xf4, 0x52, 0x90, 0x6a, 0xfe, 0x28, 0xab, 0x66, 0xc5, 0x28, 0x0c, 0xe7, 0x56, 0x80};

//******************************************************************************
//******************************************************************************
// Local Routines
//******************************************************************************
//******************************************************************************

/*************************************************************************
 * Function:        CheckForNewAttach
 *
 * Preconditions:   None
 *
 * Input:           None
 *
 * Output:          deviceAddress (global)
 *                  Updates the device address when an attach is found.
 *
 * Returns:         true if a new device has been attached.  false,
 *                  otherwise.
 *
 * Side Effects:    Prints attach message
 *
 * Overview:        This routine checks to see if a new device has been
 *                  attached.  If it has, it records the address.
 *************************************************************************/

bool CheckForNewAttach(void) {
    // Try to get the device address, if we don't have one.
    if (deviceAddress == 0) {
        GENERIC_DEVICE_ID DevID;

        DevID.vid = 0x04D8; //NO meaning: Any number is OK.
        DevID.pid = 0x000C; //NO meaning: Any number is OK.
#ifdef USB_GENERIC_SUPPORT_SERIAL_NUMBERS
        DevID.serialNumberLength = 0;
        DevID.serialNumber = NULL;
#endif

        if (USBHostGenericGetDeviceAddress(&DevID)) {
            deviceAddress = DevID.deviceAddress;
            UART2PrintString("Generic demo device attached - polled, deviceAddress=");
            UART2PutDec(deviceAddress);
            UART2PrintString("\r\n");
            return true;
        }
    }

    return false;

} // CheckForNewAttach

/*************************************************************************
 * Function:        ManageDemoState
 *
 * Preconditions:   The DemoState global variable must be initialized to
 *                  DEMO_STATE_IDLE (0).  (This occurs on reset.)
 *
 * Input:           DemoState (global)
 *                  Actions selected based value of DemoState on function
 *                  entry.
 *
 *                  deviceAddress (global)
 *                  May use device address to access device, depending on
 *                  state.
 *
 *                  DataPacket (global)
 *                  May read data from packet buffer, depending on state.
 *
 * Output:          DemoState (global)
 *                  Updates demo state as appropriate.
 *
 *                  DataPacket (global)
 *                  May cause data in the packet buffer to be updated,
 *                  depending on state.
 *
 * Returns:         None
 *
 * Side Effects:    Depend on state transition
 *
 * Overview:        This routine maintains the state of the application,
 *                  updateing global data and taking actions as necessary
 *                  to maintain the custom demo operations.
 *************************************************************************/
void ManageDemoState(void) {
    uint8_t RetVal;

    if (USBHostGenericDeviceDetached(deviceAddress) && deviceAddress != 0) {
        UART2PrintString("Generic demo device detached - polled\r\n");
        DemoState = BT_INITIALIZE;
        HciState = HCI_CMD_RESET;
        deviceAddress = 0;
    }

    switch (DemoState) {
        case BT_INITIALIZE:
            DemoState = BT_STATE_IDLE;
            break;

            /** Idle State:  Loops here until attach **/
        case BT_STATE_IDLE:
            if (CheckForNewAttach()) {
                DemoState = BT_STATE_PROCESS;
                HciState = HCI_CMD_RESET;
            }
            break;

        case BT_STATE_PROCESS:
            switch (HciState) {
                    //HCI layer***********************************************************************
                case HCI_CMD_RESET:
                    buf1[0] = 0x03;
                    buf1[1] = 0x0c;
                    buf1[2] = 0;
                    data_size = 3;
                    DemoState = BT_STATE_WRITE_CLASS;
                    HciState = HCI_CMD_RESET_END;
                    break;

                case HCI_CMD_RESET_END:
                    end_num = BL_EVENT; // When buf1[0]==end_num, reading process ends. See BT_STATE_READ_EP1
                    strcpy(message, "HCI_CMD_RESET: "); //message for BT_STATE_READ_EP1
                    DemoState = BT_STATE_READ_EP1;
                    HciState = HCI_CMD_READ_BD_ADDR;
                    break;

                    //********************************************************************************
                case HCI_CMD_READ_BD_ADDR:
                    buf1[0] = 0x09;
                    buf1[1] = 0x10;
                    buf1[2] = 0;
                    data_size = 3;
                    DemoState = BT_STATE_WRITE_CLASS;
                    HciState = HCI_CMD_READ_BD_ADDR_END;
                    break;

                case HCI_CMD_READ_BD_ADDR_END:
                    end_num = BL_EVENT;
                    strcpy(message, "HCI_CMD_READ_BD_ADDR: ");
                    DemoState = BT_STATE_READ_EP1;
                    HciState = HCI_CMD_MASK;
                    break;


                    //********************************************************************************
                case HCI_CMD_MASK:

                    for (step = 0; step < 6; step++) INI_BT_ADDRESS[step] = buf1[6 + step];

#ifdef DEBUG_MODE
                    for (data_num = 0; data_num < 6; data_num++) {
                        UART2PutHex(INI_BT_ADDRESS[data_num]);
                        UART2PutChar(' ');
                    }
                    UART2PrintString("\r\n");
#endif

                    buf1[0] = 0x01;
                    buf1[1] = 0x0c;
                    buf1[2] = 0x08;
                    buf1[3] = 0xbf; //af;//2f; + Encryption Change Event + Disconnection
                    buf1[4] = 0xff;
                    buf1[5] = 0xff;
                    buf1[6] = 0xff;
                    buf1[7] = 0xff;
                    buf1[8] = 0xff;
                    buf1[9] = 0xbf;
                    buf1[10] = 0x3d;
                    data_size = 11;
                    DemoState = BT_STATE_WRITE_CLASS;
                    HciState = HCI_CMD_MASK_END;
                    break;

                case HCI_CMD_MASK_END:
                    end_num = BL_EVENT;
                    strcpy(message, "HCI_CMD_MASK: ");
                    DemoState = BT_STATE_READ_EP1;
                    HciState = HCI_CMD_SCAN_PARAM;
                    break;

                    //********************************************************************************
                case HCI_CMD_SCAN_PARAM:
                    //LE Set Scan Parameters Command
                    buf1[0] = 0x0b;
                    buf1[1] = 0x20;
                    buf1[2] = 0x07; // PDU length
                    buf1[3] = 0x01; // Active scanning
                    buf1[4] = 0x20; //04;// LE_Scan_Interval
                    buf1[5] = 0x10;
                    buf1[6] = 0x00; // LE_Scan_Window
                    buf1[7] = 0x08;
                    buf1[8] = 0x00; // Public Device Address 
                    buf1[9] = 0x00; // Accept all advertisement packets

                    data_size = 10;
                    DemoState = BT_STATE_WRITE_CLASS;
                    HciState = HCI_CMD_SCAN_PARAM_END;
                    break;

                case HCI_CMD_SCAN_PARAM_END:
                    end_num = BL_EVENT;
                    strcpy(message, " HCI_CMD_SCAN_PARAM: ");
                    DemoState = BT_STATE_READ_EP1;
                    HciState = HCI_CMD_SCAN_ENABLE;
                    break;

                    //********************************************************************************
                case HCI_CMD_SCAN_ENABLE:
                    //Set Scan Enabl Command

                    buf1[0] = 0x0c;
                    buf1[1] = 0x20;
                    buf1[2] = 0x02; // length
                    buf1[3] = 0x01; // Scanning enabled
                    buf1[4] = 0x01; // Duplicate filtering enabled

                    data_size = 5;
                    DemoState = BT_STATE_WRITE_CLASS;
                    HciState = HCI_CMD_SCAN_ENABLE_END;
                    break;

                case HCI_CMD_SCAN_ENABLE_END:
                    end_num = BLE_EVENT;
                    strcpy(message, "HCI_CMD_SCAN_ENABLE: ");
                    DemoState = BT_STATE_READ_EP1;
                    HciState = HCI_CMD_SCAN_ENABLE_END1;
                    break;

                case HCI_CMD_SCAN_ENABLE_END1:
                    if (buf1[6] != 0x2d) {
                        HciState = HCI_CMD_SCAN_ENABLE_END;
                        break;
                    }
                    end_num = BLE_EVENT;
                    strcpy(message, "HCI_CMD_SCAN_ENABLE_END1 ");
                    DemoState = BT_STATE_READ_EP1;
                    HciState = HCI_CMD_SCAN_DISABLE;
                    break;

                    //********************************************************************************
                case HCI_CMD_SCAN_DISABLE:
                    //Set Scan Enabl Command
                    INI_ADDRESS_TYPE = buf1[5];
                    for (step = 0; step < 6; step++) RES_BT_ADDRESS[step] = buf1[step + 6];

                    buf1[0] = 0x0c;
                    buf1[1] = 0x20;
                    buf1[2] = 0x02; // length
                    buf1[3] = 0x00; // Scanning enabled
                    buf1[4] = 0x00; // Duplicate filtering enabled

                    data_size = 5;
                    DemoState = BT_STATE_WRITE_CLASS;
                    HciState = HCI_CMD_SCAN_DISABLE_END;
                    break;

                case HCI_CMD_SCAN_DISABLE_END:
                    end_num = BL_EVENT;
                    strcpy(message, "HCI_CMD_SCAN_DISABLE: ");
                    DemoState = BT_STATE_READ_EP1;
                    HciState = HCI_CMD_CLEAR_LIST;
                    break;

                    //********************************************************************************
                case HCI_CMD_CLEAR_LIST:
                    // LE Clear White List Command

                    buf1[0] = 0x10;
                    buf1[1] = 0x20;
                    buf1[2] = 0x00; // length

                    data_size = 3;
                    DemoState = BT_STATE_WRITE_CLASS;
                    HciState = HCI_CMD_CLEAR_LIST_END;
                    break;

                case HCI_CMD_CLEAR_LIST_END:
                    end_num = BL_EVENT;
                    strcpy(message, "HCI_CMD_CLEAR_LIST: ");
                    DemoState = BT_STATE_READ_EP1;
                    HciState = HCI_CMD_WRITE_LIST;
                    break;

                    //********************************************************************************
                case HCI_CMD_WRITE_LIST:
                    // LE Add Device To White List Command

                    INI_ADDRESS_TYPE = buf1[5];

                    buf1[0] = 0x11;
                    buf1[1] = 0x20;
                    buf1[2] = 0x07; // length
                    buf1[3] = INI_ADDRESS_TYPE; // Address_Type
                    for (step = 0; step < 6; step++) buf1[4 + step] = RES_BT_ADDRESS[step];

                    data_size = 10;

                    DemoState = BT_STATE_WRITE_CLASS;
                    HciState = HCI_CMD_WRITE_LIST_END;
                    break;

                case HCI_CMD_WRITE_LIST_END:
                    end_num = BL_EVENT;
                    strcpy(message, "HCI_CMD_WRITE_LIST: ");
                    DemoState = BT_STATE_READ_EP1;
                    HciState = HCI_CMD_CREATE_CONNECT;
                    break;
                    //********************************************************************************
                case HCI_CMD_CREATE_CONNECT:
                    // LE Create Connection Command

                    buf1[0] = 0x0d;
                    buf1[1] = 0x20;
                    buf1[2] = 0x19; // length
                    buf1[3] = 0x00; // LE_Scan_Interval
                    buf1[4] = 0x08;
                    buf1[5] = 0x10; // LE_Scan_Window
                    buf1[6] = 0x01; // 
                    buf1[7] = 0x01; // Initiator_Filter_Policy
                    buf1[8] = 0x00; // Peer_Address_Type

                    for (step = 0; step < 6; step++) buf1[9 + step] = RES_BT_ADDRESS[step];

                    buf1[15] = 0x00; // Own_Address_Type
                    buf1[16] = 0x06; //0x0a;// Conn_Interval_Min
                    buf1[17] = 0x00;
                    buf1[18] = 0x06; //0xa0;// Conn_Interval_Max
                    buf1[19] = 0x00;
                    buf1[20] = 0x00; // Conn_Latency
                    buf1[21] = 0x00;
                    buf1[22] = 0x90; // Supervision_Timeout
                    buf1[23] = 0x01;
                    buf1[24] = 0x00; // Minimum_CE_Length
                    buf1[25] = 0x00;
                    buf1[26] = 0x00; // Maximum_CE_Length
                    buf1[27] = 0x00;

                    data_size = 28;
                    DemoState = BT_STATE_WRITE_CLASS;
                    HciState = HCI_CMD_CREATE_CONNECT_END;
                    break;


                case HCI_CMD_CREATE_CONNECT_END:
                    end_num = BLE_EVENT;
                    strcpy(message, "HCI_CMD_CREATE_CONNECT: ");
                    DemoState = BT_STATE_READ_EP1;
                    HciState = SMP_SECURITY_REQ;
                    break;

                    //********************************************************************************
                    //********************************************************************************
                    // SMP LAYER
                    //********************************************************************************
                    //********************************************************************************

                case SMP_SECURITY_REQ:
                    strcpy(message, "SMP_SECURITY_REQ:\r\n");
                    DemoState = BT_STATE_READ_ACL_HCI;
                    HciState = SMP_PAIRING_REQ;
                    break;

                    //********************************************************************************
                case SMP_PAIRING_REQ:

                    handle[0] = buf1[4];
                    handle[1] = buf1[5];

                    buf[0] = handle[0];
                    buf[1] = 0x00;
                    buf[2] = 0x0b;
                    buf[3] = 0x00;
                    buf[4] = 0x07;
                    buf[5] = 0x00;
                    buf[6] = 0x06;
                    buf[7] = 0x00;

                    for (step = 0; step < 7; step++) buf[step + 8] = PAIR_REQ[step];

                    data_size = 15;

                    DemoState = BT_STATE_WRITE_ACL;
                    HciState = SMP_PAIR_RES;

                    break;

                case SMP_PAIR_RES:
                    strcpy(message, "SMP_PAIR_RES:\r\n");
                    DemoState = BT_STATE_READ_ACL_HCI;
                    HciState = SMP_FUNC_E;
                    break;

                    //********************************************************************************
                    //Pair config calculation
                case SMP_FUNC_E:
                    for (step = 0; step < 7; step++) PAIR_RES[step] = buf[8 + step];

                    P1[0] = INI_ADDRESS_TYPE;
                    P1[1] = RES_ADDRESS_TYPE;
                    for (step = 0; step < 7; step++) {
                        P1[2 + step] = PAIR_REQ[step];
                        P1[9 + step] = PAIR_RES[step];
                    }

                    buf1[0] = 0x17;
                    buf1[1] = 0x20;
                    buf1[2] = 0x20;
                    for (step = 0; step < 16; step++) buf1[3 + step] = 0x00;
                    for (step = 0; step < 16; step++) buf1[19 + step] = INIT_RANDOM [step] ^ P1[step];

                    data_size = 35;

#ifdef DEBUG_MODE
                    for (data_num = 0; data_num < 35; data_num++) {
                        UART2PutHex(buf1[data_num]);
                        UART2PutChar(' ');
                    }
                    UART2PrintString("\r\n");
#endif

                    DemoState = BT_STATE_WRITE_CLASS;
                    HciState = SMP_FUNC_E_END;
                    break;

                case SMP_FUNC_E_END:
                    end_num = BL_EVENT;
                    strcpy(message, "SMP_FUNC_E_END:\r\n");
                    DemoState = BT_STATE_READ_EP1;
                    HciState = SMP_FUNC_C1;
                    break;

                case SMP_FUNC_C1:

                    for (step = 0; step < 6; step++) {
                        P2[step] = RES_BT_ADDRESS[step];
                        P2[6 + step] = INI_BT_ADDRESS[step];
                    }
                    for (step = 0; step < 4; step++) P2[12 + step] = 0x00;

                    buf1[0] = 0x17;
                    buf1[1] = 0x20;
                    buf1[2] = 0x20;
                    for (step = 0; step < 16; step++) buf1[34 - step] = buf1[21 - step] ^ P2[15 - step];
                    for (step = 0; step < 16; step++) buf1[3 + step] = 0x00;
                    data_size = 35;

#ifdef DEBUG_MODE
                    for (data_num = 0; data_num < 35; data_num++) {
                        UART2PutHex(buf1[data_num]);
                        UART2PutChar(' ');
                    }
                    UART2PrintString("\r\n");
#endif

                    DemoState = BT_STATE_WRITE_CLASS;
                    HciState = SMP_FUNC_C1_END;
                    break;

                case SMP_FUNC_C1_END:
                    end_num = BL_EVENT;
                    strcpy(message, "SMP_FUNC_C1_END:\r\n");
                    DemoState = BT_STATE_READ_EP1;
                    HciState = SMP_PAIR_CONF_OUT;
                    break;

                case SMP_PAIR_CONF_OUT:

                    for (step = 0; step < 16; step++) PAIR_CONF[step] = buf1[6 + step];

#ifdef DEBUG_MODE
                    for (data_num = 0; data_num < 16; data_num++) {
                        UART2PutHex(PAIR_CONF[data_num]);
                        UART2PutChar(' ');
                    }
                    UART2PrintString("\r\n");
#endif

                    buf[1] = 0x00;
                    buf[2] = 0x15;
                    buf[3] = 0x00;
                    buf[4] = 0x11;
                    buf[5] = 0x00;
                    buf[6] = 0x06;
                    buf[7] = 0x00;
                    buf[8] = 0x03; //Pairing Confirm
                    for (step = 0; step < 16; step++) buf[9 + step] = PAIR_CONF[step];

                    data_size = 25;
                    DemoState = BT_STATE_WRITE_ACL;
                    HciState = SMP_PAIR_CONF_IN;
                    break;

                    //********************************************************************************       

                case SMP_PAIR_CONF_IN:
                    strcpy(message, "SMP_PAIR_CONF_IN:\r\n");
                    DemoState = BT_STATE_READ_ACL_HCI;
                    HciState = SMP_RANDOM_OUT;
                    break;

                case SMP_RANDOM_OUT:

                    buf[1] = 0x00;
                    buf[2] = 0x15;
                    buf[3] = 0x00;
                    buf[4] = 0x11;
                    buf[5] = 0x00;
                    buf[6] = 0x06;
                    buf[7] = 0x00;
                    buf[8] = 0x04; //
                    for (step = 0; step < 16; step++) buf[9 + step] = INIT_RANDOM[step];

                    data_size = 25;
                    DemoState = BT_STATE_WRITE_ACL;
                    HciState = SMP_RANDOM_IN;
                    break;

                case SMP_RANDOM_IN:
                    strcpy(message, "SMP_RANDOM_IN:\r\n");
                    DemoState = BT_STATE_READ_ACL_HCI;
                    HciState = SMP_FUNC_S1;
                    break;


                    //********************************************************************************
                    //STK = s1(TK, Srand, Mrand)
                case SMP_FUNC_S1:

                    for (step = 0; step < 16; step++) PAIR_RANDOM[step] = buf[9 + step];

#ifdef DEBUG_MODE
                    for (data_num = 0; data_num < 16; data_num++) {
                        UART2PutHex(PAIR_RANDOM[data_num]);
                        UART2PutChar(' ');
                    }
                    UART2PrintString("\r\n");
#endif

                    buf1[0] = 0x17;
                    buf1[1] = 0x20;
                    buf1[2] = 0x20;
                    for (step = 0; step < 16; step++) buf1[3 + step] = 0x00;
                    for (step = 0; step < 8; step++) {
                        buf1[19 + step] = INIT_RANDOM[step];
                        buf1[27 + step] = PAIR_RANDOM[step];
                    }
                    data_size = 35;

#ifdef DEBUG_MODE
                    for (data_num = 0; data_num < 35; data_num++) {
                        UART2PutHex(buf1[data_num]);
                        UART2PutChar(' ');
                    }
                    UART2PrintString("\r\n");
#endif

                    DemoState = BT_STATE_WRITE_CLASS;
                    HciState = SMP_FUNC_S1_END;
                    break;

                case SMP_FUNC_S1_END:
                    end_num = BL_EVENT;
                    strcpy(message, "SMP_FUNC_S1_END:\r\n");
                    DemoState = BT_STATE_READ_EP1;
                    HciState = SMP_LTK_REQ;
                    break;

                    //********************************************************************************
                case SMP_LTK_REQ:

                    for (step = 0; step < 16; step++) STK[step] = buf1[6 + step];

                    buf1[0] = 0x19;
                    buf1[1] = 0x20;
                    buf1[2] = 0x1c;
                    buf1[3] = buf[0]; //handle
                    buf1[4] = 0x00;

                    buf1[5] = 0x00; //random number
                    buf1[6] = 0x00;
                    buf1[7] = 0x00;
                    buf1[8] = 0x00;
                    buf1[9] = 0x00;
                    buf1[10] = 0x00;
                    buf1[11] = 0x00;
                    buf1[12] = 0x00;

                    buf1[13] = 0x00; //EDIV
                    buf1[14] = 0x00;

                    for (step = 0; step < 16; step++) buf1[15 + step] = STK[step];

                    data_size = 31;

#ifdef DEBUG_MODE
                    for (data_num = 0; data_num < 31; data_num++) {
                        UART2PutHex(buf1[data_num]);
                        UART2PutChar(' ');
                    }
                    UART2PrintString("\r\n");
#endif

                    DemoState = BT_STATE_WRITE_CLASS;
                    HciState = SMP_LTK_RES;
                    break;

                case SMP_LTK_RES:
                    end_num = ENCRYPT_EVENT;
                    strcpy(message, "SMP_LTK_RES:\r\n");
                    DemoState = BT_STATE_READ_EP1;
                    HciState = SMP_ENCRYP_INFO_IN;
                    break;

                case SMP_ENCRYP_INFO_IN:
                    strcpy(message, "SMP_ENCRYP_INFO_IN:\r\n");
                    DemoState = BT_STATE_READ_ACL_HCI;
                    HciState = SMP_RANDEDIV_IN;
                    break;

                case SMP_RANDEDIV_IN:
                    strcpy(message, "SMP_RANDEDIV_IN:\r\n");
                    DemoState = BT_STATE_READ_ACL_HCI;
                    HciState = SMP_ENCRYP_INFO_OUT;
                    break;

                case SMP_ENCRYP_INFO_OUT:

                    buf[1] = 0x00;
                    buf[2] = 0x15;
                    buf[3] = 0x00;
                    buf[4] = 0x11;
                    buf[5] = 0x00;
                    buf[6] = 0x06;
                    buf[7] = 0x00;
                    buf[8] = 0x08;

                    for (step = 0; step < 16; step++) buf[9 + step] = IRK[step];

                    data_size = 25;

                    DemoState = BT_STATE_WRITE_ACL;
                    HciState = SMP_RANDEDIV_OUT;
                    break;

                case SMP_RANDEDIV_OUT:

                    buf[1] = 0x00;
                    buf[2] = 0x0c;
                    buf[3] = 0x00;
                    buf[4] = 0x08;
                    buf[5] = 0x00;
                    buf[6] = 0x06;
                    buf[7] = 0x00;
                    buf[8] = 0x09;
                    buf[9] = 0x00;

                    for (step = 0; step < 6; step++) buf[10 + step] = INI_BT_ADDRESS[step];

                    data_size = 16;

                    DemoState = BT_STATE_WRITE_ACL;
                    HciState = GATT_MTU_REQ;
                    break;

                    //********************************************************************************
                    //********************************************************************************
                    // GATT LAYER
                    //********************************************************************************
                    //********************************************************************************
                case GATT_MTU_REQ:

                    buf[1] = 0x00;
                    buf[2] = 0x07;
                    buf[3] = 0x00;
                    buf[4] = 0x03;
                    buf[5] = 0x00;
                    buf[6] = 0x04;
                    buf[7] = 0x00;
                    buf[8] = 0x02;
                    buf[9] = 0x0f;
                    buf[10] = 0x02;

                    data_size = 11;

                    DemoState = BT_STATE_WRITE_ACL;
                    HciState = GATT_MTU_RES;
                    break;

                case GATT_MTU_RES:
                    strcpy(message, "GATT_MTU_RES:\r\n");
                    DemoState = BT_STATE_READ_ACL_HCI;
                    HciState = GATT_GEN_ATT_CCC;
                    break;

                case GATT_GEN_ATT_CCC:

                    buf[1] = 0x00;
                    buf[2] = 0x09;
                    buf[3] = 0x00;
                    buf[4] = 0x05;
                    buf[5] = 0x00;
                    buf[6] = 0x04;
                    buf[7] = 0x00;
                    buf[8] = 0x12;
                    buf[9] = 0x04;
                    buf[10] = 0x00;
                    buf[11] = 0x02;
                    buf[12] = 0x00;

                    data_size = 13;

                    DemoState = BT_STATE_WRITE_ACL;
                    HciState = GATT_GEN_ATT_CCC_RES;
                    break;

                case GATT_GEN_ATT_CCC_RES:

                    strcpy(message, "GATT_GEN_ATT_CCC_RES:\r\n");
                    DemoState = BT_STATE_READ_ACL_HCI;
                    HciState = GATT_GEN_HID1_CCC;
                    break;

                case GATT_GEN_HID1_CCC:

                    buf[1] = 0x00;
                    buf[2] = 0x09;
                    buf[3] = 0x00;
                    buf[4] = 0x05;
                    buf[5] = 0x00;
                    buf[6] = 0x04;
                    buf[7] = 0x00;
                    buf[8] = 0x12;
                    buf[9] = 0x14;
                    buf[10] = 0x00;
                    buf[11] = 0x01;
                    buf[12] = 0x00;

                    data_size = 13;

                    DemoState = BT_STATE_WRITE_ACL;
                    HciState = GATT_GEN_ATT_HID1_RES;
                    break;

                case GATT_GEN_ATT_HID1_RES:

                    strcpy(message, "GATT_GEN_ATT_HID1_RES:\r\n");
                    DemoState = BT_STATE_READ_ACL_HCI;
                    HciState = GATT_GEN_HID2_CCC;
                    break;

                case GATT_GEN_HID2_CCC:

                    buf[1] = 0x00;
                    buf[2] = 0x09;
                    buf[3] = 0x00;
                    buf[4] = 0x05;
                    buf[5] = 0x00;
                    buf[6] = 0x04;
                    buf[7] = 0x00;
                    buf[8] = 0x12;
                    buf[9] = 0x18;
                    buf[10] = 0x00;
                    buf[11] = 0x01;
                    buf[12] = 0x00;

                    data_size = 13;

                    DemoState = BT_STATE_WRITE_ACL;
                    HciState = GATT_GEN_ATT_HID2_RES;
                    break;

                case GATT_GEN_ATT_HID2_RES:
                    strcpy(message, "GATT_GEN_ATT_HID2_RES:\r\n");
                    DemoState = BT_STATE_READ_ACL_HCI;
                    HciState = READ_HID_DATA;
                    LATBbits.LATB15 = 1; //LED ON
                    break;

                    //********************************************************************************
                    //********************************************************************************
                    // HOGP LAYER
                    //********************************************************************************
                    //********************************************************************************

                case READ_HID_DATA:
                    strcpy(message, "READ_HID_DATA:\r\n");
                    DemoState = BT_STATE_READ_ACL_HCI;
                    HciState = READ_HID_DATA1;
                    break;

                case READ_HID_DATA1:
                    strcpy(message, "READ_HID_DATA1:\r\n");

                    //  LED ON-OFF START: If you don't need, remove from here to //  LED ON-OFF END            
                    if (buf[9] == 0x17 && buf[12] == 0x28) {
                        HID17_flag = 1;
                        LED_On(); //LED ON
                    } else if (buf[9] == 0x13 && buf[11] == 0x01) {
                        if (HID17_flag == 1) HID17_flag = 0;
                        else LED_Off(); //LED OFF
                    }
                    //  LED ON-OFF END 

                    if (buf[6] == 0x05 && buf[8] == 0x12) HciState = READ_HID_DATA2;
                    else HciState = READ_HID_DATA;
                    break;

                case READ_HID_DATA2:

                    buf[1] = 0x00;
                    buf[2] = 0x0a;
                    buf[3] = 0x00;
                    buf[4] = 0x06;
                    buf[5] = 0x00;
                    buf[6] = 0x05;
                    buf[7] = 0x00;
                    buf[8] = 0x13;
                    buf[9] = 0x01;
                    buf[10] = 0x02;
                    buf[11] = 0x00;
                    buf[12] = 0x00;
                    buf[13] = 0x00;

                    data_size = 14;

                    DemoState = BT_STATE_WRITE_ACL;
                    HciState = READ_HID_DATA;
                    break;

                    //********************************************************************************
                    //********************************************************************************
                    //DISCONNECTION
                case L2CAP_DISCONNECT_REQ:
                    buf1[0] = 0x06;
                    buf1[1] = 0x04;
                    buf1[2] = 0x03;
                    buf1[3] = handle[0];
                    buf1[4] = handle[1];
                    buf1[5] = 0x13;

                    DemoState = BT_STATE_WRITE_CLASS;
                    HciState = L2CAP_DISCONNECT_REQ_END;
                    break;

                case L2CAP_DISCONNECT_REQ_END:

                    UART2PrintString("HCI_DISCONNECT !:\r\n");

                    DelayMs(100);
                    LED_Off(); //LED OFF
                    HciState = HCI_CMD_SCAN_ENABLE;
                    break;

                    //********************************************************************************
                    //********************************************************************************

                case PROG_END:
                    break;

            }
            break;


            //********************************************************************************
            //?READ-WRITE FUNCTIONS
            //********************************************************************************
            //WRITE ENDPOINT 0
        case BT_STATE_WRITE_CLASS:
            if (!USBHostGenericTxIsBusy(deviceAddress)) {
                if ((RetVal = USBHostGenericClassRequest(deviceAddress, buf1, buf1[2] + 3)) == USB_SUCCESS) {
                    //UART2PrintString( "HCI COMMAND SENT\r\n" );	
                    DemoState = BT_STATE_PROCESS;
                } else {
                    UART2PrintString("Write Class Error !\r\n");
                }
            }
            break;


            //READ ENDPOINT 1
        case BT_STATE_READ_EP1:
            if (!USBHostGenericRx1IsBusy(deviceAddress)) {
                if ((RetVal = USBHostGenericRead(deviceAddress, buf1, MAX_NO_OF_BYTES)) == USB_SUCCESS) {
                    DemoState = BT_STATE_READ_CLASS_WAITING;
                    //UART2PrintString( "READ EP1\r\n" );
                } else {
                    UART2PrintString("Device Read Error !\r\n");
                }
            }
            break;

        case BT_STATE_READ_CLASS_WAITING:
            if (!USBHostGenericRx1IsBusy(deviceAddress)) {

#ifdef DEBUG_MODE
                UART2PrintString(message);
                for (data_num = 0; data_num < buf1[1] + 2; data_num++) {
                    UART2PutHex(buf1[data_num]);
                    UART2PutChar(' ');
                }
                UART2PrintString("\r\n");
#endif
                if (buf1[0] != end_num) {
                    DemoState = BT_STATE_READ_EP1;
                    break;
                }



                DemoState = BT_STATE_PROCESS;
            }
            break;


            //WRITE ENDPOINT 2
        case BT_STATE_WRITE_ACL:

#ifdef DEBUG_MODE
            for (data_num = 0; data_num < buf[2] + 4; data_num++) {
                UART2PutHex(buf[data_num]);
                UART2PutChar(' ');
            }
            UART2PrintString("\r\n");
#endif

            if (!USBHostGenericTxIsBusy(deviceAddress)) {
                if ((RetVal = USBHostGenericAclWrite(deviceAddress, buf, buf[2] + 4)) == USB_SUCCESS) {
                    //UART2PrintString( "HCI COMMAND SENT\r\n" );	
                    DemoState = BT_STATE_PROCESS;
                } else {
                    UART2PrintString("Write Acl Error !\r\n");
                }
            }
            break;

            //READ ENDPOINT 2 & 1 
        case BT_STATE_READ_ACL_HCI:
            if (EP2busy_flag == 1) {
                DemoState = BT_STATE_READ_ACL_WAITING;
                break;
            }
            if (!USBHostGenericRx2IsBusy(deviceAddress)) {
                if ((RetVal = USBHostGenericAclRead(deviceAddress, buf, MAX_NO_OF_BYTES)) == USB_SUCCESS) {
                    DemoState = BT_STATE_READ_ACL_WAITING;
                } else {
                    UART2PrintString("Read Acl Error !\r\n");
                }
            }
            break;

        case BT_STATE_READ_ACL_WAITING:
            if (!USBHostGenericRx2IsBusy(deviceAddress)) {

#ifdef DEBUG_MODE
                UART2PrintString(message);
                for (data_num = 0; data_num < buf[2] + 4; data_num++) {
                    UART2PutHex(buf[data_num]);
                    UART2PutChar(' ');
                }
                UART2PrintString("\r\n");
#endif
                EP2busy_flag = 0;
                DemoState = BT_STATE_READ_HCI;
            } else if (hid_flag != 3) {
                EP2busy_flag = 1;
                DemoState = BT_STATE_READ_HCI;
            }
            break;

        case BT_STATE_READ_HCI:
            if (EP2busy_flag == 0) buf1[0] = 0xff;
            if (!USBHostGenericRx1IsBusy(deviceAddress)) {
                if ((RetVal = USBHostGenericRead(deviceAddress, buf1, MAX_NO_OF_BYTES)) == USB_SUCCESS) {
                    DemoState = BT_STATE_READ_HCI_WAITING;
                    //UART2PrintString( "READ EP1\r\n" );
                } else {
                    UART2PrintString("Device Read Error !\r\n");
                }
            }
            break;

        case BT_STATE_READ_HCI_WAITING:
            if (!USBHostGenericRx1IsBusy(deviceAddress)) {

#ifdef DEBUG_MODE
                UART2PrintString("HCI: ");
                for (data_num = 0; data_num < buf1[1] + 2; data_num++) {
                    UART2PutHex(buf1[data_num]);
                    UART2PutChar(' ');
                }
                UART2PrintString("\r\n");
#endif

                if (buf1[0] == DISCONNECT) {
                    HciState = L2CAP_DISCONNECT_REQ;
                    DemoState = BT_STATE_PROCESS;
                    break;
                }// following code is not needed in principle, but it is needed in practice in debug mode
                else if (buf1[0] == BLE_EVENT && buf1[1] == 0x0d) {
                    HciState = HCI_CMD_RESET;
                    DemoState = BT_STATE_PROCESS;
                    break;
                }

                if (EP2busy_flag == 0) {
                    if (buf1[0] == 0xff) {
                        DemoState = BT_STATE_PROCESS;
                    } else {
                        DemoState = BT_STATE_READ_HCI;
                    }
                } else {
                    //				if(buf1[0]==DISCONNECT){HciState = L2CAP_DISCONNECT_REQ;DemoState = BT_STATE_PROCESS;}
                    //              else if(buf1[0]==BLE_EVENT && buf1[1]==0x0d){HciState = LTK_REQ_REPLY;DemoState = BT_STATE_PROCESS;}
                    //				else {DemoState = BT_STATE_READ_ACL_WAITING;}
                    DemoState = BT_STATE_READ_ACL_WAITING;
                }
            }
            break;

            //********************************************************************************

            /** Error state:  Hold here until detached **/
        case BT_STATE_ERROR: // To Do: Flash LEDs
            break;

        default:
            DemoState = BT_INITIALIZE;
            break;
    }

    DelayMs(1); // 1ms delay

} // ManageDemoState

/*
void __attribute__((interrupt, no_auto_psv)) _INT1Interrupt(void) {

    DemoState = BT_STATE_PROCESS;
    HciState = L2CAP_DISCONNECT_REQ;
    strcpy(message, "DISCONNECT:\r\n");
    IFS1bits.INT1IF = 0;
}
 */

//******************************************************************************
//******************************************************************************
// USB Support Functions
//******************************************************************************
//******************************************************************************

/*************************************************************************
 * Function:        USB_ApplicationEventHandler
 *
 * Preconditions:   The USB must be initialized.
 *
 * Input:           event       Identifies the bus event that occured
 *
 *                  data        Pointer to event-specific data
 *
 *                  size        Size of the event-specific data
 *
 * Output:          deviceAddress (global)
 *                  Updates device address when an attach or detach occurs.
 *
 *                  DemoState (global)
 *                  Updates the demo state as appropriate when events occur.
 *
 * Returns:         true if the event was handled, false if not
 *
 * Side Effects:    Event-specific actions have been taken.
 *
 * Overview:        This routine is called by the Host layer or client
 *                  driver to notify the application of events that occur.
 *                  If the event is recognized, it is handled and the
 *                  routine returns true.  Otherwise, it is ignored (or
 *                  just "sniffed" and the routine returns false.
 *************************************************************************/

bool USB_ApplicationEventHandler(uint8_t address, USB_EVENT event, void *data, uint32_t size) {
#ifdef USB_GENERIC_SUPPORT_SERIAL_NUMBERS
    uint8_t i;
#endif

    // Handle specific events.
    switch ((int) event) {
        case EVENT_GENERIC_ATTACH:
            if (size == sizeof (GENERIC_DEVICE_ID)) {
                deviceAddress = ((GENERIC_DEVICE_ID *) data)->deviceAddress;
                DemoState = BT_STATE_PROCESS;
                HciState = HCI_CMD_RESET; //YTS !!!!!!!!!!!!!!!!!
                UART2PrintString("Generic demo device attached - event, deviceAddress=");
                UART2PutDec(deviceAddress);
                UART2PrintString("\r\n");
#ifdef USB_GENERIC_SUPPORT_SERIAL_NUMBERS
                for (i = 1; i < ((GENERIC_DEVICE_ID *) data)->serialNumberLength; i++) {
                    UART2PutChar(((GENERIC_DEVICE_ID *) data)->serialNumber[i]);
                }
#endif
                UART2PrintString("\r\n");
                return true;
            }
            break;

        case EVENT_GENERIC_DETACH:
            deviceAddress = 0;
            DemoState = BT_INITIALIZE; //YTS
            UART2PrintString("Generic demo device detached - event\r\n");
            return true;

        case EVENT_GENERIC_TX_DONE: // The main state machine will poll the driver.
        case EVENT_GENERIC_RX1_DONE: //YTS
        case EVENT_GENERIC_RX2_DONE: //YTS
            return true;

        case EVENT_VBUS_REQUEST_POWER:
            // We'll let anything attach.
            return true;

        case EVENT_VBUS_RELEASE_POWER:
            // We aren't keeping track of power.
            return true;

        case EVENT_HUB_ATTACH:
            UART2PrintString("\r\n***** USB Error - hubs are not supported *****\r\n");
            return true;
            break;

        case EVENT_UNSUPPORTED_DEVICE:
            UART2PrintString("\r\n***** USB Error - device is not supported *****\r\n");
            return true;
            break;

        case EVENT_CANNOT_ENUMERATE:
            UART2PrintString("\r\n***** USB Error - cannot enumerate device *****\r\n");
            return true;
            break;

        case EVENT_CLIENT_INIT_ERROR:
            UART2PrintString("\r\n***** USB Error - client driver initialization error *****\r\n");
            return true;
            break;

        case EVENT_OUT_OF_MEMORY:
            UART2PrintString("\r\n***** USB Error - out of heap memory *****\r\n");
            return true;
            break;

        case EVENT_UNSPECIFIED_ERROR: // This should never be generated.
            UART2PrintString("\r\n***** USB Error - unspecified *****\r\n");
            return true;
            break;

        case EVENT_SUSPEND:
        case EVENT_DETACH:
        case EVENT_RESUME:
        case EVENT_BUS_ERROR:
            return true;
            break;

        default:
            break;
    }

    return false;

} // USB_ApplicationEventHandler


//******************************************************************************
//******************************************************************************
// Main
//******************************************************************************
//******************************************************************************

/*************************************************************************
 * Function:        main
 *
 * Preconditions:   None
 *
 * Input:           None
 *
 * Output:          None
 *
 * Returns:         Never exits
 *
 * Side Effects:    Runs the application
 *
 * Overview:        This is the USB Custom Demo Application's main entry
 *                  point.
 *************************************************************************/

int main(void) {

    // Initialize the processor and peripherals.
    SYSTEM_Initialize();

    // Set Default demo state
    DemoState = BT_INITIALIZE;
    HciState = HCI_CMD_RESET;

    LED_Off();

    // Main Processing Loop
    while (1) {
        // This demo does not check for overcurrent conditions.  See the
        // USB Host - Data Logger for an example of overcurrent detection
        // with the PIC24F and the USB PICtail Plus.

        // Maintain USB Host State
        USBHostTasks();
        // Maintain Demo Application State
        ManageDemoState();
    }

    return 0;

} // main


/*************************************************************************
 * EOF main.c
 */


